@@ -10,10 +10,9 @@
 #include <private/qqmlengine_p.h>
 #include <private/qv4scopedvalue_p.h>
 #include <private/qv4jscall_p.h>
-#include "qv4runtime_p.h"
-#include "qv4objectiterator_p.h"
 #include <private/qqmlmetatype_p.h>
 #include <private/qqmltype_p_p.h>
+#include <private/qqmlvaluetypewrapper_p.h>
 
 #include <algorithm>
 
@@ -23,18 +22,38 @@ namespace QV4 {
 
 DEFINE_OBJECT_VTABLE(Sequence);
 
+static ReturnedValue doGetIndexed(const Sequence *s, qsizetype index) {
+    QV4::Scope scope(s->engine());
+
+    Heap::ReferenceObject::Flags flags =
+            Heap::ReferenceObject::EnforcesLocation;
+    if (s->d()->typePrivate()->extraData.ld->canSetValueAtIndex())
+        flags |= Heap::ReferenceObject::CanWriteBack;
+    if (Sequence::valueMetaType(s->d()) == QMetaType::fromType<QVariant>())
+        flags |= Heap::ReferenceObject::IsVariant;
+
+    QV4::ScopedValue v(scope, scope.engine->fromVariant(
+                           s->at(index), s->d(), index, flags));
+    if (QQmlValueTypeWrapper *ref = v->as<QQmlValueTypeWrapper>()) {
+        if (CppStackFrame *frame = scope.engine->currentStackFrame)
+            ref->d()->setLocation(frame->v4Function, frame->statementNumber());
+        // No need to read the reference. at() has done that already.
+    }
+    return v->asReturnedValue();
+}
+
 static const QMetaSequence *metaSequence(const Heap::Sequence *p)
 {
-    return p->typePrivate->extraData.ld;
+    return p->typePrivate()->extraData.ld;
 }
 
 template<typename Compare>
 void sortSequence(Sequence *sequence, const Compare &compare)
 {
-    const auto *p = sequence->d();
+    auto *p = sequence->d();
     const auto *m = metaSequence(p);
 
-    QSequentialIterable iterable(*m, p->typePrivate->listId, p->container);
+    QSequentialIterable iterable(*m, p->typePrivate()->listId, p->storagePointer());
     if (iterable.canRandomAccessIterate()) {
         std::sort(QSequentialIterable::RandomAccessIterator(iterable.mutableBegin()),
                   QSequentialIterable::RandomAccessIterator(iterable.mutableEnd()),
@@ -71,11 +90,8 @@ struct SequenceOwnPropertyKeyIterator : ObjectOwnPropertyKeyIterator
     {
         const Sequence *s = static_cast<const Sequence *>(o);
 
-        if (s->d()->isReference) {
-            if (!s->d()->object)
-                return ObjectOwnPropertyKeyIterator::next(o, pd, attrs);
-            s->loadReference();
-        }
+        if (s->d()->isReference() && !s->loadReference())
+            return PropertyKey::invalid();
 
         const qsizetype size = s->size();
         if (size > 0 && qIsAtMostSizetypeLimit(arrayIndex, size - 1)) {
@@ -84,7 +100,7 @@ struct SequenceOwnPropertyKeyIterator : ObjectOwnPropertyKeyIterator
             if (attrs)
                 *attrs = QV4::Attr_Data;
             if (pd)
-                pd->value = s->engine()->fromVariant(s->at(qsizetype(index)));
+                pd->value = doGetIndexed(s, index);
             return PropertyKey::fromArrayIndex(index);
         }
 
@@ -134,17 +150,13 @@ struct SequenceDefaultCompareFunctor
 
 void Heap::Sequence::init(const QQmlType &qmlType, const void *container)
 {
-    Object::init();
+    ReferenceObject::init(nullptr, -1, NoFlag);
 
     Q_ASSERT(qmlType.isSequentialContainer());
-    typePrivate = qmlType.priv();
-    QQmlType::refHandle(typePrivate);
+    m_typePrivate = qmlType.priv();
+    QQmlType::refHandle(m_typePrivate);
 
-    this->container = typePrivate->listId.create(container);
-    propertyIndex = -1;
-    isReference = false;
-    isReadOnly = false;
-    object.init();
+    m_container = m_typePrivate->listId.create(container);
 
     QV4::Scope scope(internalClass->engine);
     QV4::Scoped<QV4::Sequence> o(scope, this);
@@ -152,53 +164,103 @@ void Heap::Sequence::init(const QQmlType &qmlType, const void *container)
 }
 
 void Heap::Sequence::init(
-        QObject *object, int propertyIndex, const QQmlType &qmlType, bool readOnly)
+        const QQmlType &qmlType, const void *container,
+        Heap::Object *object, int propertyIndex, Heap::ReferenceObject::Flags flags)
 {
-    Object::init();
+    ReferenceObject::init(object, propertyIndex, flags);
 
     Q_ASSERT(qmlType.isSequentialContainer());
-    typePrivate = qmlType.priv();
-    QQmlType::refHandle(typePrivate);
-    container = QMetaType(typePrivate->listId).create();
-    this->propertyIndex = propertyIndex;
-    isReference = true;
-    this->isReadOnly = readOnly;
-    this->object.init(object);
+    m_typePrivate = qmlType.priv();
+    QQmlType::refHandle(m_typePrivate);
     QV4::Scope scope(internalClass->engine);
     QV4::Scoped<QV4::Sequence> o(scope, this);
     o->setArrayType(Heap::ArrayData::Custom);
-    o->loadReference();
+    if (CppStackFrame *frame = scope.engine->currentStackFrame)
+        setLocation(frame->v4Function, frame->statementNumber());
+    if (container)
+        m_container = QMetaType(m_typePrivate->listId).create(container);
+    else if (flags & EnforcesLocation)
+        QV4::ReferenceObject::readReference(this);
+}
+
+Heap::Sequence *Heap::Sequence::detached() const
+{
+    return internalClass->engine->memoryManager->allocate<QV4::Sequence>(
+                QQmlType(m_typePrivate), m_container);
 }
 
 void Heap::Sequence::destroy()
 {
-    typePrivate->listId.destroy(container);
-    QQmlType::derefHandle(typePrivate);
-    object.destroy();
-    Object::destroy();
+    if (m_container)
+        m_typePrivate->listId.destroy(m_container);
+    QQmlType::derefHandle(m_typePrivate);
+    ReferenceObject::destroy();
+}
+
+void *Heap::Sequence::storagePointer()
+{
+    if (!m_container)
+        m_container = m_typePrivate->listId.create();
+    return m_container;
+}
+
+bool Heap::Sequence::setVariant(const QVariant &variant)
+{
+    const QMetaType variantReferenceType = variant.metaType();
+    if (variantReferenceType != m_typePrivate->listId) {
+        // This is a stale reference. That is, the property has been
+        // overwritten with a different type in the meantime.
+        // We need to modify this reference to the updated type, if
+        // possible, or return false if it is not a sequence.
+        const QQmlType newType = QQmlMetaType::qmlListType(variantReferenceType);
+        if (newType.isSequentialContainer()) {
+            if (m_container)
+                m_typePrivate->listId.destroy(m_container);
+            QQmlType::derefHandle(m_typePrivate);
+            m_typePrivate = newType.priv();
+            QQmlType::refHandle(m_typePrivate);
+            m_container = m_typePrivate->listId.create(variant.constData());
+            return true;
+        } else {
+            return false;
+        }
+    }
+    if (m_container) {
+        variantReferenceType.destruct(m_container);
+        variantReferenceType.construct(m_container, variant.constData());
+    } else {
+        m_container = variantReferenceType.create(variant.constData());
+    }
+    return true;
+}
+QVariant Heap::Sequence::toVariant() const
+{
+    return QVariant(m_typePrivate->listId, m_container);
 }
 
 const QMetaType Sequence::valueMetaType(const Heap::Sequence *p)
 {
-    return p->typePrivate->typeId;
+    return p->typePrivate()->typeId;
 }
 
 qsizetype Sequence::size() const
 {
     const auto *p = d();
-    return metaSequence(p)->size(p->container);
+    Q_ASSERT(p->storagePointer()); // Must readReference() before
+    return metaSequence(p)->size(p->storagePointer());
 }
 
 QVariant Sequence::at(qsizetype index) const
 {
     const auto *p = d();
+    Q_ASSERT(p->storagePointer()); // Must readReference() before
     const QMetaType v = valueMetaType(p);
     QVariant result;
     if (v == QMetaType::fromType<QVariant>()) {
-        metaSequence(p)->valueAtIndex(p->container, index, &result);
+        metaSequence(p)->valueAtIndex(p->storagePointer(), index, &result);
     } else {
         result = QVariant(v);
-        metaSequence(p)->valueAtIndex(p->container, index, result.data());
+        metaSequence(p)->valueAtIndex(p->storagePointer(), index, result.data());
     }
     return result;
 }
@@ -221,18 +283,18 @@ void convertAndDo(const QVariant &item, const QMetaType v, Action action)
 
 void Sequence::append(const QVariant &item)
 {
-    const Heap::Sequence *p = d();
+    Heap::Sequence *p = d();
     convertAndDo(item, valueMetaType(p), [p](const void *data) {
-        metaSequence(p)->addValueAtEnd(p->container, data);
+        metaSequence(p)->addValueAtEnd(p->storagePointer(), data);
     });
 }
 
 void Sequence::append(qsizetype num, const QVariant &item)
 {
-    const Heap::Sequence *p = d();
+    Heap::Sequence *p = d();
     convertAndDo(item, valueMetaType(p), [p, num](const void *data) {
         const QMetaSequence *m = metaSequence(p);
-        void *container = p->container;
+        void *container = p->storagePointer();
         for (qsizetype i = 0; i < num; ++i)
             m->addValueAtEnd(container, data);
     });
@@ -240,50 +302,39 @@ void Sequence::append(qsizetype num, const QVariant &item)
 
 void Sequence::replace(qsizetype index, const QVariant &item)
 {
-    const Heap::Sequence *p = d();
+    Heap::Sequence *p = d();
     convertAndDo(item, valueMetaType(p), [p, index](const void *data) {
-        metaSequence(p)->setValueAtIndex(p->container, index, data);
+        metaSequence(p)->setValueAtIndex(p->storagePointer(), index, data);
     });
 }
 
 void Sequence::removeLast(qsizetype num)
 {
-    const auto *p = d();
+    auto *p = d();
     const auto *m = metaSequence(p);
 
     if (m->canEraseRangeAtIterator() && m->hasRandomAccessIterator() && num > 1) {
-        void *i = m->end(p->container);
+        void *i = m->end(p->storagePointer());
         m->advanceIterator(i, -num);
-        void *j = m->end(p->container);
-        m->eraseRangeAtIterator(p->container, i, j);
+        void *j = m->end(p->storagePointer());
+        m->eraseRangeAtIterator(p->storagePointer(), i, j);
         m->destroyIterator(i);
         m->destroyIterator(j);
     } else {
         for (int i = 0; i < num; ++i)
-            m->removeValueAtEnd(p->container);
+            m->removeValueAtEnd(p->storagePointer());
     }
 }
 
-QVariant Sequence::toVariant() const
-{
-    const auto *p = d();
-    return QVariant(p->typePrivate->listId, p->container);
-}
-
 ReturnedValue Sequence::containerGetIndexed(qsizetype index, bool *hasProperty) const
 {
-    if (d()->isReference) {
-        if (!d()->object) {
-            if (hasProperty)
-                *hasProperty = false;
-            return Encode::undefined();
-        }
-        loadReference();
-    }
+    if (d()->isReference() && !loadReference())
+        return Encode::undefined();
+
     if (index >= 0 && index < size()) {
         if (hasProperty)
             *hasProperty = true;
-        return engine()->fromVariant(at(index));
+        return doGetIndexed(this, index);
     }
     if (hasProperty)
         *hasProperty = false;
@@ -295,20 +346,17 @@ bool Sequence::containerPutIndexed(qsizetype index, const Value &value)
     if (internalClass()->engine->hasException)
         return false;
 
-    if (d()->isReadOnly) {
+    if (d()->isReadOnly()) {
         engine()->throwTypeError(QLatin1String("Cannot insert into a readonly container"));
         return false;
     }
 
-    if (d()->isReference) {
-        if (!d()->object)
-            return false;
-        loadReference();
-    }
+    if (d()->isReference() && !loadReference())
+        return false;
 
     const qsizetype count = size();
     const QMetaType valueType = valueMetaType(d());
-    const QVariant element = engine()->toVariant(value, valueType, false);
+    const QVariant element = ExecutionEngine::toVariant(value, valueType, false);
 
     if (index < 0)
         return false;
@@ -325,7 +373,7 @@ bool Sequence::containerPutIndexed(qsizetype index, const Value &value)
         append(element);
     }
 
-    if (d()->isReference)
+    if (d()->object())
         storeReference();
     return true;
 }
@@ -338,14 +386,10 @@ SequenceOwnPropertyKeyIterator *containerOwnPropertyKeys(const Object *m, Value
 
 bool Sequence::containerDeleteIndexedProperty(qsizetype index)
 {
-    if (d()->isReadOnly)
+    if (d()->isReadOnly())
+        return false;
+    if (d()->isReference() && !loadReference())
         return false;
-    if (d()->isReference) {
-        if (!d()->object)
-            return false;
-        loadReference();
-    }
-
     if (index < 0 || index >= size())
         return false;
 
@@ -353,7 +397,7 @@ bool Sequence::containerDeleteIndexedProperty(qsizetype index)
     /* but we cannot, so we insert a default-value instead. */
     replace(index, QVariant());
 
-    if (d()->isReference)
+    if (d()->object())
         storeReference();
 
     return true;
@@ -366,9 +410,10 @@ bool Sequence::containerIsEqualTo(Managed *other)
     Sequence *otherSequence = other->as<Sequence>();
     if (!otherSequence)
         return false;
-    if (d()->isReference && otherSequence->d()->isReference) {
-        return d()->object == otherSequence->d()->object && d()->propertyIndex == otherSequence->d()->propertyIndex;
-    } else if (!d()->isReference && !otherSequence->d()->isReference) {
+    if (d()->object() && otherSequence->d()->object()) {
+        return d()->object() == otherSequence->d()->object()
+                && d()->property() == otherSequence->d()->property();
+    } else if (!d()->object() && !otherSequence->d()->object()) {
         return this == otherSequence;
     }
     return false;
@@ -376,44 +421,36 @@ bool Sequence::containerIsEqualTo(Managed *other)
 
 bool Sequence::sort(const FunctionObject *f, const Value *, const Value *argv, int argc)
 {
-    if (d()->isReadOnly)
+    if (d()->isReadOnly())
+        return false;
+    if (d()->isReference() && !loadReference())
         return false;
-    if (d()->isReference) {
-        if (!d()->object)
-            return false;
-        loadReference();
-    }
 
     if (argc == 1 && argv[0].as<FunctionObject>())
         sortSequence(this, SequenceCompareFunctor(f->engine(), argv[0]));
     else
         sortSequence(this, SequenceDefaultCompareFunctor());
 
-    if (d()->isReference)
+    if (d()->object())
         storeReference();
 
     return true;
 }
 
 void *Sequence::getRawContainerPtr() const
-{ return d()->container; }
+{ return d()->storagePointer(); }
 
-void Sequence::loadReference() const
+bool Sequence::loadReference() const
 {
-    Q_ASSERT(d()->object);
-    Q_ASSERT(d()->isReference);
-    void *a[] = { d()->container, nullptr };
-    QMetaObject::metacall(d()->object, QMetaObject::ReadProperty, d()->propertyIndex, a);
+    Q_ASSERT(d()->object());
+    // If locations are enforced we only read once
+    return d()->enforcesLocation() || QV4::ReferenceObject::readReference(d());
 }
 
-void Sequence::storeReference()
+bool Sequence::storeReference()
 {
-    Q_ASSERT(d()->object);
-    Q_ASSERT(d()->isReference);
-    int status = -1;
-    QQmlPropertyData::WriteFlags flags = QQmlPropertyData::DontRemoveBinding;
-    void *a[] = { d()->container, nullptr, &status, &flags };
-    QMetaObject::metacall(d()->object, QMetaObject::WriteProperty, d()->propertyIndex, a);
+    Q_ASSERT(d()->object());
+    return d()->isAttachedToProperty() && QV4::ReferenceObject::writeBack(d());
 }
 
 ReturnedValue Sequence::virtualGet(const Managed *that, PropertyKey id, const Value *receiver, bool *hasProperty)
@@ -430,6 +467,14 @@ ReturnedValue Sequence::virtualGet(const Managed *that, PropertyKey id, const Va
     return Object::virtualGet(that, id, receiver, hasProperty);
 }
 
+qint64 Sequence::virtualGetLength(const Managed *m)
+{
+    const Sequence *s = static_cast<const Sequence *>(m);
+    if (s->d()->isReference() && !s->loadReference())
+        return 0;
+    return s->size();
+}
+
 bool Sequence::virtualPut(Managed *that, PropertyKey id, const Value &value, Value *receiver)
 {
     if (id.isArrayIndex()) {
@@ -466,6 +511,43 @@ OwnPropertyKeyIterator *Sequence::virtualOwnPropertyKeys(const Object *m, Value
     return containerOwnPropertyKeys(m, target);
 }
 
+int Sequence::virtualMetacall(Object *object, QMetaObject::Call call, int index, void **a)
+{
+    Sequence *sequence = static_cast<Sequence *>(object);
+    Q_ASSERT(sequence);
+
+    switch (call) {
+    case QMetaObject::ReadProperty: {
+        const QMetaType valueType = valueMetaType(sequence->d());
+        if (sequence->d()->isReference() && !sequence->loadReference())
+            return 0;
+        const QMetaSequence *metaSequence = sequence->d()->typePrivate()->extraData.ld;
+        if (metaSequence->valueMetaType() != valueType)
+            return 0; // value metatype is not what the caller expects anymore.
+
+        const void *storagePointer = sequence->d()->storagePointer();
+        if (index < 0 || index >= metaSequence->size(storagePointer))
+            return 0;
+        metaSequence->valueAtIndex(storagePointer, index, a[0]);
+        break;
+    }
+    case QMetaObject::WriteProperty: {
+        void *storagePointer = sequence->d()->storagePointer();
+        const QMetaSequence *metaSequence = sequence->d()->typePrivate()->extraData.ld;
+        if (index < 0 || index >= metaSequence->size(storagePointer))
+            return 0;
+        metaSequence->setValueAtIndex(storagePointer, index, a[0]);
+        if (sequence->d()->isReference())
+            sequence->storeReference();
+        break;
+    }
+    default:
+        return 0; // not supported
+    }
+
+    return -1;
+}
+
 static QV4::ReturnedValue method_get_length(const FunctionObject *b, const Value *thisObject, const Value *, int)
 {
     QV4::Scope scope(b);
@@ -473,11 +555,8 @@ static QV4::ReturnedValue method_get_length(const FunctionObject *b, const Value
     if (!This)
         THROW_TYPE_ERROR();
 
-    if (This->d()->isReference) {
-        if (!This->d()->object)
-            RETURN_RESULT(Encode(0));
-        This->loadReference();
-    }
+    if (This->d()->isReference() && !This->loadReference())
+        return Encode::undefined();
 
     const qsizetype size = This->size();
     if (qIsAtMostUintLimit(size))
@@ -500,17 +579,14 @@ static QV4::ReturnedValue method_set_length(const FunctionObject *f, const Value
         RETURN_UNDEFINED();
     }
 
-    if (This->d()->isReadOnly)
+    if (This->d()->isReadOnly())
         THROW_TYPE_ERROR();
 
     const qsizetype newCount = qsizetype(argv0);
 
     /* Read the sequence from the QObject property if we're a reference */
-    if (This->d()->isReference) {
-        if (!This->d()->object)
-            RETURN_UNDEFINED();
-        This->loadReference();
-    }
+    if (This->d()->isReference() && !This->loadReference())
+        RETURN_UNDEFINED();
 
     /* Determine whether we need to modify the sequence */
     const qsizetype count = This->size();
@@ -530,10 +606,8 @@ static QV4::ReturnedValue method_set_length(const FunctionObject *f, const Value
     }
 
     /* write back if required. */
-    if (This->d()->isReference) {
-        /* write back.  already checked that object is non-null, so skip that check here. */
+    if (This->d()->object())
         This->storeReference();
-    }
 
     RETURN_UNDEFINED();
 }
@@ -554,7 +628,7 @@ ReturnedValue SequencePrototype::method_sort(const FunctionObject *b, const Valu
 {
     Scope scope(b);
     QV4::ScopedObject o(scope, thisObject);
-    if (!o || !o->isListType())
+    if (!o || !o->isV4SequenceType())
         THROW_TYPE_ERROR();
 
     if (argc >= 2)
@@ -569,10 +643,9 @@ ReturnedValue SequencePrototype::method_sort(const FunctionObject *b, const Valu
 }
 
 ReturnedValue SequencePrototype::newSequence(
-        QV4::ExecutionEngine *engine, QMetaType sequenceType, QObject *object,
-        int propertyIndex,  bool readOnly, bool *succeeded)
+        QV4::ExecutionEngine *engine, QMetaType sequenceType, const void *data,
+        Heap::Object *object, int propertyIndex, Heap::ReferenceObject::Flags flags)
 {
-    QV4::Scope scope(engine);
     // This function is called when the property is a QObject Q_PROPERTY of
     // the given sequence type.  Internally we store a sequence
     // (as well as object ptr + property index for updated-read and write-back)
@@ -580,55 +653,54 @@ ReturnedValue SequencePrototype::newSequence(
 
     const QQmlType qmlType = QQmlMetaType::qmlListType(sequenceType);
     if (qmlType.isSequentialContainer()) {
-        *succeeded = true;
-        QV4::ScopedObject obj(scope, engine->memoryManager->allocate<Sequence>(
-                                  object, propertyIndex, qmlType, readOnly));
-        return obj.asReturnedValue();
+        return engine->memoryManager->allocate<Sequence>(
+                    qmlType, data,  object, propertyIndex, flags)->asReturnedValue();
     }
 
-    *succeeded = false;
     return Encode::undefined();
 }
 
-ReturnedValue SequencePrototype::fromVariant(
-        QV4::ExecutionEngine *engine, const QVariant &v, bool *succeeded)
+ReturnedValue SequencePrototype::fromVariant(QV4::ExecutionEngine *engine, const QVariant &v)
 {
-    return fromData(engine, v.metaType(), v.constData(), succeeded);
+    return fromData(engine, v.metaType(), v.constData());
 }
 
-ReturnedValue SequencePrototype::fromData(ExecutionEngine *engine, QMetaType type, const void *data, bool *succeeded)
+ReturnedValue SequencePrototype::fromData(ExecutionEngine *engine, QMetaType type, const void *data)
 {
-    QV4::Scope scope(engine);
     // This function is called when assigning a sequence value to a normal JS var
     // in a JS block.  Internally, we store a sequence of the specified type.
     // Access and mutation is extremely fast since it will not need to modify any
     // QObject property.
 
     const QQmlType qmlType = QQmlMetaType::qmlListType(type);
-    if (qmlType.isSequentialContainer()) {
-        *succeeded = true;
-        QV4::ScopedObject obj(scope, engine->memoryManager->allocate<Sequence>(qmlType, data));
-        return obj.asReturnedValue();
-    }
+    if (qmlType.isSequentialContainer())
+        return engine->memoryManager->allocate<Sequence>(qmlType, data)->asReturnedValue();
 
-    *succeeded = false;
     return Encode::undefined();
 }
 
 QVariant SequencePrototype::toVariant(const Sequence *object)
 {
-    Q_ASSERT(object->isListType());
-    return object->toVariant();
+    Q_ASSERT(object->isV4SequenceType());
+    const auto *p = object->d();
+
+    // Note: For historical reasons, we ignore the result of loadReference()
+    //       here. This allows us to retain sequences whose objects have vaninshed
+    //       as "var" properties. It comes at the price of potentially returning
+    //       outdated data. This is the behavior sequences have always shown.
+    if (p->isReference())
+        object->loadReference();
+    if (!p->hasData())
+        return QVariant();
+
+    return QVariant(p->typePrivate()->listId, p->storagePointer());
 }
 
-QVariant SequencePrototype::toVariant(const QV4::Value &array, QMetaType typeHint, bool *succeeded)
+QVariant SequencePrototype::toVariant(const QV4::Value &array, QMetaType typeHint)
 {
-    *succeeded = true;
-
-    if (!array.as<ArrayObject>()) {
-        *succeeded = false;
+    if (!array.as<ArrayObject>())
         return QVariant();
-    }
+
     QV4::Scope scope(array.as<Object>()->engine());
     QV4::ScopedArrayObject a(scope, array);
 
@@ -645,17 +717,25 @@ QVariant SequencePrototype::toVariant(const QV4::Value &array, QMetaType typeHin
         QV4::ScopedValue v(scope);
         for (quint32 i = 0; i < quint32(length); ++i) {
             const QMetaType valueMetaType = priv->typeId;
-            QVariant variant = scope.engine->toVariant(a->get(i), valueMetaType, false);
+            QVariant variant = ExecutionEngine::toVariant(a->get(i), valueMetaType, false);
             if (valueMetaType == QMetaType::fromType<QVariant>()) {
                 meta->addValueAtEnd(result.data(), &variant);
             } else {
                 const QMetaType originalType = variant.metaType();
-                if (originalType != valueMetaType && !variant.convert(valueMetaType)) {
-                    qWarning() << QLatin1String(
-                                      "Could not convert array value at position %1 from %2 to %3")
-                                  .arg(QString::number(i), QString::fromUtf8(originalType.name()),
-                                       QString::fromUtf8(valueMetaType.name()));
-                    variant = QVariant(valueMetaType);
+                if (originalType != valueMetaType) {
+                    const QVariant converted = QQmlValueTypeProvider::createValueType(
+                                variant, valueMetaType);
+                    if (converted.isValid()) {
+                        variant = converted;
+                    } else if (!variant.convert(valueMetaType)) {
+                        qWarning().noquote()
+                                << QLatin1String("Could not convert array value "
+                                                 "at position %1 from %2 to %3")
+                                   .arg(QString::number(i),
+                                        QString::fromUtf8(originalType.name()),
+                                        QString::fromUtf8(valueMetaType.name()));
+                        variant = QVariant(valueMetaType);
+                    }
                 }
                 meta->addValueAtEnd(result.data(), variant.constData());
             }
@@ -663,20 +743,19 @@ QVariant SequencePrototype::toVariant(const QV4::Value &array, QMetaType typeHin
         return result;
     }
 
-    *succeeded = false;
     return QVariant();
 }
 
 void *SequencePrototype::getRawContainerPtr(const Sequence *object, QMetaType typeHint)
 {
-    if (object->d()->typePrivate->listId == typeHint)
+    if (object->d()->typePrivate()->listId == typeHint)
         return object->getRawContainerPtr();
     return nullptr;
 }
 
 QMetaType SequencePrototype::metaTypeForSequence(const Sequence *object)
 {
-    return object->d()->typePrivate->listId;
+    return object->d()->typePrivate()->listId;
 }
 
 } // namespace QV4
