@@ -35,7 +35,7 @@ static QQmlDirParser createQmldirParserForFile(const QString &filename)
 }
 
 void QQmlJSImporter::readQmltypes(
-        const QString &filename, QHash<QString, QQmlJSExportedScope> *objects,
+        const QString &filename, QList<QQmlJSExportedScope> *objects,
         QList<QQmlDirParser::Import> *dependencies)
 {
     const QFileInfo fileInfo(filename);
@@ -58,7 +58,15 @@ void QQmlJSImporter::readQmltypes(
     }
 
     QFile file(filename);
-    file.open(QFile::ReadOnly);
+    if (!file.open(QFile::ReadOnly)) {
+        m_warnings.append({
+                              QStringLiteral("QML types file cannot be opened: ") + filename,
+                              QtWarningMsg,
+                              QQmlJS::SourceLocation()
+                          });
+        return;
+    }
+
     QQmlJSTypeDescriptionReader reader { filename, QString::fromUtf8(file.readAll()) };
     QStringList dependencyStrings;
     auto succ = reader(objects, &dependencyStrings);
@@ -124,7 +132,6 @@ static bool isComposite(const QQmlJSScope::ConstPtr &scope)
 QQmlJSImporter::QQmlJSImporter(const QStringList &importPaths, QQmlJSResourceFileMapper *mapper,
                                bool useOptionalImports)
     : m_importPaths(importPaths),
-      m_builtins({}),
       m_mapper(mapper),
       m_useOptionalImports(useOptionalImports),
       m_createImportVisitor([](const QQmlJSScope::Ptr &target, QQmlJSImporter *importer,
@@ -141,11 +148,9 @@ QQmlJSImporter::Import QQmlJSImporter::readQmldir(const QString &path)
     Import result;
     auto reader = createQmldirParserForFile(path + SlashQmldir);
     result.name = reader.typeNamespace();
-#if QT_VERSION >= QT_VERSION_CHECK(6, 4, 0)
-    // #if required for standalone DOM build against Qt6.3
+
     result.isStaticModule = reader.isStaticModule();
     result.isSystemModule = reader.isSystemModule();
-#endif
     result.imports.append(reader.imports());
     result.dependencies.append(reader.dependencies());
 
@@ -199,7 +204,7 @@ QQmlJSImporter::Import QQmlJSImporter::readQmldir(const QString &path)
                                reader.typeNamespace(), it.key(), it->version, QTypeRevision()));
     }
     for (auto it = qmlComponents.begin(), end = qmlComponents.end(); it != end; ++it)
-        result.objects.insert(it.key(), it.value());
+        result.objects.append(it.value());
 
     const auto scripts = reader.scripts();
     for (const auto &script : scripts) {
@@ -225,7 +230,7 @@ QQmlJSImporter::Import QQmlJSImporter::readDirectory(const QString &directory)
             for (const auto &entry : resources) {
                 const QString name = QFileInfo(entry.resourcePath).baseName();
                 if (name.front().isUpper()) {
-                    import.objects.insert(name, {
+                    import.objects.append({
                         localFile2ScopeTree(entry.filePath),
                         { QQmlJSScope::Export(QString(), name, QTypeRevision(), QTypeRevision()) }
                     });
@@ -245,12 +250,21 @@ QQmlJSImporter::Import QQmlJSImporter::readDirectory(const QString &directory)
         QDir::NoFilter
     };
     while (it.hasNext()) {
-        it.next();
-        if (!it.fileName().front().isUpper())
-            continue; // Non-uppercase names cannot be imported anyway.
+        QString name = it.nextFileInfo().completeBaseName();
 
-        const QString name = QFileInfo(it.filePath()).baseName();
-        import.objects.insert(name, {
+        // Non-uppercase names cannot be imported anyway.
+        if (!name.front().isUpper())
+            continue;
+
+        // .ui.qml is fine
+        if (name.endsWith(u".ui"))
+            name = name.chopped(3);
+
+        // Names with dots in them cannot be imported either.
+        if (name.contains(u'.'))
+            continue;
+
+        import.objects.append({
                 localFile2ScopeTree(it.filePath()),
                 { QQmlJSScope::Export(QString(), name, QTypeRevision(), QTypeRevision()) }
         });
@@ -275,11 +289,9 @@ void QQmlJSImporter::importDependencies(const QQmlJSImporter::Import &import,
             if (!m_useOptionalImports) {
                 continue;
             }
-#if (QT_VERSION >= QT_VERSION_CHECK(6, 4, 0))
-            // #if required for standalone DOM build against Qt6.3
+
             if (!(import.flags & QQmlDirParser::Import::OptionalDefault))
                 continue;
-#endif
         }
 
         importHelper(import.module, types, isDependency ? QString() : prefix,
@@ -298,8 +310,14 @@ void QQmlJSImporter::importDependencies(const QQmlJSImporter::Import &import,
 static bool isVersionAllowed(const QQmlJSScope::Export &exportEntry,
                              const QQmlJSScope::Import &importDescription)
 {
-    return !importDescription.version().isValid()
-            || exportEntry.version() <= importDescription.version();
+    const QTypeRevision importVersion = importDescription.version();
+    const QTypeRevision exportVersion = exportEntry.version();
+    if (!importVersion.hasMajorVersion())
+        return true;
+    if (importVersion.majorVersion() != exportVersion.majorVersion())
+        return false;
+    return !importVersion.hasMinorVersion()
+            || exportVersion.minorVersion() <= importVersion.minorVersion();
 }
 
 void QQmlJSImporter::processImport(const QQmlJSScope::Import &importDescription,
@@ -329,8 +347,8 @@ void QQmlJSImporter::processImport(const QQmlJSScope::Import &importDescription,
             if (!bestExport.isValid() || valExport.version() > bestExport.version())
                 bestExport = valExport;
 
-            const auto it = types->qmlNames.find(qmlName);
-            if (it != types->qmlNames.end()) {
+            const auto it = types->qmlNames.types().find(qmlName);
+            if (it != types->qmlNames.types().end()) {
 
                 // The same set of exports can declare the same name multiple times for different
                 // versions. That's the common thing and we would just continue here when we hit
@@ -370,7 +388,7 @@ void QQmlJSImporter::processImport(const QQmlJSScope::Import &importDescription,
                     });
 
                     // Invalidate the type. We don't know which one to use.
-                    it->scope = QQmlJSScope::ConstPtr();
+                    types->qmlNames.clearType(qmlName);
                     continue;
                 }
                 case HigherVersion:
@@ -378,27 +396,28 @@ void QQmlJSImporter::processImport(const QQmlJSScope::Import &importDescription,
                 }
             }
 
-            types->qmlNames.insert(qmlName, { val.scope, valExport.version() });
+            types->qmlNames.setType(qmlName, { val.scope, valExport.version() });
             seenExports[qmlName].append(valExport);
         }
 
         const QTypeRevision bestRevision = bestExport.isValid()
                 ? bestExport.revision()
                 : QTypeRevision::zero();
-        types->cppNames.insert(cppName, { val.scope, bestRevision });
+        types->cppNames.setType(cppName, { val.scope, bestRevision });
 
         const QTypeRevision bestVersion = bestExport.isValid()
                 ? bestExport.version()
                 : QTypeRevision::zero();
-        types->qmlNames.insert(prefixedName(internalPrefix, cppName), { val.scope, bestVersion });
+        types->qmlNames.setType(prefixedName(internalPrefix, cppName), { val.scope, bestVersion });
     };
 
+    // Empty type means "this is the prefix"
     if (!importDescription.prefix().isEmpty())
-        types->qmlNames.insert(importDescription.prefix(), {}); // Empty type means "this is the prefix"
+        types->qmlNames.setType(importDescription.prefix(), {});
 
     // Add a marker to show that this module has been imported
     if (!importDescription.isDependency())
-        types->qmlNames.insert(prefixedName(modulePrefix, importDescription.name()), {});
+        types->qmlNames.setType(prefixedName(modulePrefix, importDescription.name()), {});
 
     if (!importDescription.isDependency()) {
         if (import.isStaticModule)
@@ -415,18 +434,16 @@ void QQmlJSImporter::processImport(const QQmlJSScope::Import &importDescription,
     }
 
     // add objects
-    for (auto it = import.objects.begin(); it != import.objects.end(); ++it) {
-        const auto &val = it.value();
-
+    for (const auto &val : import.objects) {
         const QString cppName = isComposite(val.scope)
                 ? prefixedName(anonPrefix, internalName(val.scope))
                 : internalName(val.scope);
 
         if (val.exports.isEmpty()) {
             // Insert an unresolvable dummy name
-            types->qmlNames.insert(
+            types->qmlNames.setType(
                         prefixedName(internalPrefix, cppName), { val.scope, QTypeRevision() });
-            types->cppNames.insert(cppName, { val.scope, QTypeRevision() });
+            types->cppNames.setType(cppName, { val.scope, QTypeRevision() });
         } else {
             insertExports(val, cppName);
         }
@@ -454,22 +471,22 @@ void QQmlJSImporter::processImport(const QQmlJSScope::Import &importDescription,
     */
 
     QQmlJSImporter::AvailableTypes tempTypes(builtinImportHelper().cppNames);
-    tempTypes.cppNames.insert(types->cppNames);
+    tempTypes.cppNames.addTypes(types->cppNames);
 
     // At present, there are corner cases that couldn't be resolved in a single
     // pass of resolveTypes() (e.g. QQmlEasingEnums::Type). However, such cases
     // only happen when enumerations are involved, thus the strategy is to
     // resolve enumerations (which can potentially create new child scopes)
     // before resolving the type fully
-    const QQmlJSScope::ConstPtr intType = tempTypes.cppNames.value(u"int"_s).scope;
+    const QQmlJSScope::ConstPtr arrayType = tempTypes.cppNames.type(u"Array"_s).scope;
     for (auto it = import.objects.begin(); it != import.objects.end(); ++it) {
-        if (!it->scope.factory())
-            QQmlJSScope::resolveEnums(it->scope, intType);
+        if (!it->scope.factory()) {
+            QQmlJSScope::resolveEnums(it->scope, tempTypes.cppNames);
+            QQmlJSScope::resolveList(it->scope, arrayType);
+        }
     }
 
-    for (auto it = import.objects.begin(); it != import.objects.end(); ++it) {
-        const auto &val = it.value();
-
+    for (const auto &val : std::as_const(import.objects)) {
         // Otherwise we have already done it in localFile2ScopeTree()
         if (!val.scope.factory() && val.scope->baseType().isNull()) {
 
@@ -501,8 +518,10 @@ QQmlJSImporter::ImportedTypes QQmlJSImporter::importBuiltins()
 
 QQmlJSImporter::AvailableTypes QQmlJSImporter::builtinImportHelper()
 {
-    if (!m_builtins.qmlNames.isEmpty() || !m_builtins.cppNames.isEmpty())
-        return m_builtins;
+    if (m_builtins)
+        return *m_builtins;
+
+    AvailableTypes builtins(ImportedTypes(ImportedTypes::INTERNAL, {}, {}));
 
     Import result;
     result.name = QStringLiteral("QML");
@@ -511,13 +530,13 @@ QQmlJSImporter::AvailableTypes QQmlJSImporter::builtinImportHelper()
                                   QStringLiteral("jsroot.qmltypes") };
     const auto importBuiltins = [&](const QStringList &imports) {
         for (auto const &dir : imports) {
-            QDirIterator it { dir, qmltypesFiles, QDir::NoFilter, QDirIterator::Subdirectories };
+            QDirIterator it { dir, qmltypesFiles, QDir::NoFilter };
             while (it.hasNext() && !qmltypesFiles.isEmpty()) {
                 readQmltypes(it.next(), &result.objects, &result.dependencies);
                 qmltypesFiles.removeOne(it.fileName());
             }
             setQualifiedNamesOn(result);
-            importDependencies(result, &m_builtins);
+            importDependencies(result, &builtins);
 
             if (qmltypesFiles.isEmpty())
                 return;
@@ -540,9 +559,33 @@ QQmlJSImporter::AvailableTypes QQmlJSImporter::builtinImportHelper()
     // otherwise
     const QQmlJSScope::Import builtinImport(
                 QString(), QStringLiteral("QML"), QTypeRevision::fromVersion(1, 0), false, true);
-    processImport(builtinImport, result, &m_builtins);
 
-    return m_builtins;
+    QQmlJSScope::ConstPtr intType;
+    QQmlJSScope::ConstPtr arrayType;
+
+    for (const QQmlJSExportedScope &exported : result.objects) {
+        if (exported.scope->internalName() == u"int"_s) {
+            intType = exported.scope;
+            if (!arrayType.isNull())
+                break;
+        } else if (exported.scope->internalName() == u"Array"_s) {
+            arrayType = exported.scope;
+            if (!intType.isNull())
+                break;
+        }
+    }
+
+    Q_ASSERT(intType);
+    Q_ASSERT(arrayType);
+
+    m_builtins = AvailableTypes(
+                ImportedTypes(ImportedTypes::INTERNAL, builtins.cppNames.types(), arrayType));
+    m_builtins->qmlNames
+            = ImportedTypes(ImportedTypes::QML, builtins.qmlNames.types(), arrayType);
+
+    processImport(builtinImport, result, &(*m_builtins));
+
+    return *m_builtins;
 }
 
 /*!
@@ -550,8 +593,6 @@ QQmlJSImporter::AvailableTypes QQmlJSImporter::builtinImportHelper()
  */
 void QQmlJSImporter::importQmldirs(const QStringList &qmldirFiles)
 {
-    AvailableTypes types(builtinImportHelper().cppNames);
-
     for (const auto &file : qmldirFiles) {
         Import result;
         QString qmldirName;
@@ -603,8 +644,10 @@ QQmlJSImporter::ImportedTypes QQmlJSImporter::importModule(const QString &module
 
     // If we imported a system module add all builtin QML types
     if (result.hasSystemModule) {
-        for (const QString &name : builtins.qmlNames.keys())
-            result.qmlNames.insert(prefixedName(prefix, name), builtins.qmlNames[name]);
+        for (auto nameIt = builtins.qmlNames.types().keyBegin(),
+                  end = builtins.qmlNames.types().keyEnd();
+             nameIt != end; ++nameIt)
+            result.qmlNames.setType(prefixedName(prefix, *nameIt), builtins.qmlNames.type(*nameIt));
     }
 
     if (staticModuleList)
@@ -636,13 +679,13 @@ bool QQmlJSImporter::importHelper(const QString &module, AvailableTypes *types,
 
         const auto &cacheEntry = m_cachedImportTypes[cacheKey];
 
-        types->cppNames.insert(cacheEntry->cppNames);
+        types->cppNames.addTypes(cacheEntry->cppNames);
         types->staticModules << cacheEntry->staticModules;
-        types->hasSystemModule = cacheEntry->hasSystemModule;
+        types->hasSystemModule |= cacheEntry->hasSystemModule;
 
         // No need to import qml names for dependencies
         if (!isDependency)
-            types->qmlNames.insert(cacheEntry->qmlNames);
+            types->qmlNames.addTypes(cacheEntry->qmlNames);
 
         return true;
     };
@@ -655,8 +698,9 @@ bool QQmlJSImporter::importHelper(const QString &module, AvailableTypes *types,
     if (getTypesFromCache())
         return true;
 
-    auto cacheTypes =
-            QSharedPointer<QQmlJSImporter::AvailableTypes>(new QQmlJSImporter::AvailableTypes({}));
+    auto cacheTypes = QSharedPointer<QQmlJSImporter::AvailableTypes>(
+                new QQmlJSImporter::AvailableTypes(
+                    ImportedTypes(ImportedTypes::INTERNAL, {}, types->cppNames.arrayType())));
     m_cachedImportTypes[cacheKey] = cacheTypes;
 
     const QPair<QString, QTypeRevision> importId { module, version };
@@ -770,7 +814,10 @@ QQmlJSScope::Ptr QQmlJSImporter::importFile(const QString &file)
 QQmlJSImporter::ImportedTypes QQmlJSImporter::importDirectory(
         const QString &directory, const QString &prefix)
 {
-    QQmlJSImporter::AvailableTypes types({});
+    const AvailableTypes builtins = builtinImportHelper();
+    QQmlJSImporter::AvailableTypes types(
+                ImportedTypes(
+                    ImportedTypes::INTERNAL, {}, builtins.cppNames.arrayType()));
     importHelper(directory, &types, prefix, QTypeRevision(), false, true);
     return types.qmlNames;
 }
@@ -786,9 +833,17 @@ void QQmlJSImporter::setImportPaths(const QStringList &importPaths)
     // Luckily this doesn't apply to m_seenQmldirFiles
 }
 
+void QQmlJSImporter::clearCache()
+{
+    m_seenImports.clear();
+    m_cachedImportTypes.clear();
+    m_seenQmldirFiles.clear();
+    m_importedFiles.clear();
+}
+
 QQmlJSScope::ConstPtr QQmlJSImporter::jsGlobalObject() const
 {
-    return m_builtins.cppNames[u"GlobalObject"_s].scope;
+    return m_builtins->cppNames.type(u"GlobalObject"_s).scope;
 }
 
 void QQmlJSImporter::setQualifiedNamesOn(const Import &import)
