@@ -2,10 +2,12 @@
 // This code is governed by the BSD license found in the LICENSE file.
 
 /*---
-es5id: 15.2.3.11-4-15
-description: Object.isSealed returns false for all built-in objects (Date)
+es5id: 15.2.3.11-4-23
+description: >
+    Object.isSealed returns false for all built-in objects
+    (ReferenceError)
 ---*/
 
-var b = Object.isSealed(Date);
+var b = Object.isSealed(ReferenceError);
 
 assert.sameValue(b, false, 'b');
