@@ -86,27 +86,7 @@ static QString findClassName(const QJSManagedValue &value)
     if (QV4::ScopedValue scoped(scope, asManaged(value)); scoped->isManaged())
         return scoped->managed()->vtable()->className;
 
-    Q_UNREACHABLE();
-    return QString();
-}
-
-static QJsonValue createPublicNumberMethod(const QString &name, int numArguments)
-{
-    QJsonObject method;
-    method.insert(QStringLiteral("access"), QStringLiteral("public"));
-    method.insert(QStringLiteral("name"), name);
-    method.insert(QStringLiteral("returnType"), QStringLiteral("number"));
-    QJsonArray arguments;
-    // arguments look nicer when 1-based than 0-based.
-    for (int i = 1; i <= numArguments; ++i) {
-        QJsonObject argument;
-        argument.insert(QStringLiteral("name"), QStringLiteral("arg%1").arg(i));
-        argument.insert(QStringLiteral("type"), QStringLiteral("number"));
-        arguments.append(argument);
-    }
-    method.insert(QStringLiteral("arguments"), arguments);
-    return method;
-
+    Q_UNREACHABLE_RETURN(QString());
 }
 
 static QString buildClass(const QJSManagedValue &value, QJsonArray *classes,
@@ -249,15 +229,6 @@ static QString buildClass(const QJSManagedValue &value, QJsonArray *classes,
                 methodObject.insert(
                             QStringLiteral("returnType"),
                             buildConstructor(prop, classes, seen, ctorName, &constructed));
-            } else if (name == QStringLiteral("Math")) {
-                // All methods of Math return numbers when given numbers.
-                // We model this as overloads of the generic methods.
-                methods.append(createPublicNumberMethod(info.name, formalParams));
-
-                // If there is no formal parameter, it doesn't matter what you pass.
-                // You always get number back.
-                if (formalParams == 0)
-                    continue;
             }
 
             QJsonArray arguments;
